------------------------------------------------------------------------------------------------
-- Starcraft 2 Model (3ds->M3) Exporter version 0.06
-- by NiNtoxicated (madyavic@gmail.com)
-- Exports M3 models from 3ds max via maxscript
-- Created for 3ds max 2010, but should be backwards compatible
-- Based on reworked code from my World of Warcraft M2 importer/exporter
--
-- Currently exports:
-- - Geometry
-- - Materials
-- - Bounding Box
--
-- Setting up an exportable model:
-- 1. Use one mesh of 'editable_mesh' type only.
-- 2. Apply one 'Multimaterial' material to it only.
-- 3. Export!
--
-- Big thanks to: 
-- Teal (starcraft.incgamers.com) - PHP M3 parser
-- Volcore (http://volcore.limbicsoft.com/)  - Helped figure out vertex flags
-- Witchsong (http://code.google.com/p/libm3/) - Awesome M3 library and for help on sequence data
-- Sixen (http://www.sc2mapster.com/) - Providing a good central resource for SC2 tools
-- der_Ton (http://www.doom3world.org/) - Great work on MD5 format which is similar to M3
-- ufoZ - One of the original people to reverse engineer the M2 format and provide a good maxscript importer
-- 			from which my importer/exporters were originally based
-- Check out http://code.google.com/p/libm3/ for more indepth detail about the M3 format
-- Head to http://www.sc2mapster.com/ for more mod tools and Starcraft 2 content
--
-- If you use any of this code in your own scripts, please credit
-- where it is due.
--
-- Things to do:
-- 1. Clean up messy code
-- 2. Implement 3dsMax GUI properly
-- 3. Add good animation support
-- 4. Improve code functionality
-- 5. Improve processing times
-- 6. Add more feature support
-- 7. Improve error catching and error support
--
------------------------------------------------------------------------------------------------
--
-- Version History:
-- 0.06 -- May 1st 2010
-- Basic UI added
-- Some basic error catching added
--
-- 0.04 -- April 30th 2010
-- Adapted to new MD34 format
-- Basic geometry exporting implemented
-- Bounding sphere exporting added
--
-- 0.01 -- April 12th 2010
-- Initial version of M3 export script
--
------------------------------------------------------------------------------------------------

-- ********************
--  SCRIPT DECLARES
-- ********************
-- globals
global head, cmat, mwrite, aread, bstream, cbones
global scrStart, scrEnd -- For script timeStamps
global twrite		= #()
global cmesh		= #()
global usedMats	= #()

-- Switches
global g_flipuvy = true -- Flips uv.y coords

-- Misc. variables
global g_mapchannel = 1

-- Temp. write file
--global m2emod = "G:\\3dsmax7\\meshes\\ArcaneGolem.m2"
global texPath = "Assets/Textures/"

-- Function Declares
fn WriteRef bitStream data reftype flags:0 = ()
fn WriteArray bitStream data flags:#uint32 reftype:#null abool:true  = ()

-- *******************
--  M3 STRUCTURES
-- ******************* 
struct M3E_SeqData
(
	tag, ablock
)

struct M3E_Sphere
(
	emin = [0,0,0], emax = [0,0,0], rad = 0, flags = 0,
	
	fn GetBound mesh =
	(
		sph = M3E_Sphere()
		sph.emin = mesh.min
		sph.emax = mesh.max
		local vec3 = mesh.max - mesh.center
		local rad = 0
		for i = 1 to 3 do
		(
			-- set radius to highest x, y or z
			if (vec3[i] > rad) then rad = vec3[i]
		)
		-- increase radius slightly...
		rad *= 1.25
		sph.rad = rad
		
		return sph
	)
)

struct matrix
(
	row1 = [1,0,0,0], row2 = [0,1,0,0], row3 = [0,0,1,0], row4 = [0,0,0,1]
)

-- ***********
--  HELPERS
-- ***********
-- Liberally stolen from my M2 script
fn echo msg =
(
	format "%\n" (msg) to:listener
)
	
fn LongToString num=
(
	local str = ""
	for i = 1 to 4 do
	(
		str += bit.intAsChar (bit.and num 0xff)
		-- num = bit.shift num -8
		num /= 256
	)
	str
)

fn ShortToFloat &val = -- Short to Float func
(
	if (val > 0) then (val -= 32767) else (val += 32767)
	val /= 32767.0
)

fn M3E_Convert_Time t =
(
	t / (1000 / useFPS)
)

fn M3E_Reset_Globals =
(
	head = mwrite = bstream = cmesh = undefined
	cmat = #()
	aread = #()
	twrite = #()
	cbones = #()
	usedMats = #()
	gc()
)

-- Writing Funcs
-- Helpers
fn M3E_Convert_Time t =
(
	t * (1000 / useFPS)	
)

fn WriteVec bitStream vec flags bflags:#null =
(
	case flags of
	(
		#v3dshort: (writeShort bitStream vec.x #unsigned; writeShort bitStream vec.y #unsigned; writeShort bitStream vec.z #unsigned;)
		#v3dlong: (writeLong bitStream vec.x #unsigned; writeLong bitStream vec.y #unsigned; writeLong bitStream vec.z #unsigned;)
		#v3dbyte: (writeByte bitStream vec.x #unsigned; writeByte bitStream vec.y #unsigned; writeByte bitStream vec.z #unsigned;)
		#v4d: (writeFloat bitStream vec.x; writeFloat bitStream vec.y; writeFloat bitStream vec.z; writeFloat bitStream vec.w;)
		#v4dshort: (writeShort bitStream vec.x #signed; writeShort bitStream vec.y #signed; writeShort bitStream vec.z #signed; writeShort bitStream vec.w #signed;)
		#v4dbyte: (writeByte bitStream vec.x; writeByte bitStream vec.y; writeByte bitStream vec.z; writeByte bitStream vec.w;)
		#v3d: (writeFloat bitStream vec.x; writeFloat bitStream vec.y; writeFloat bitStream vec.z;)
		#matrix: (WriteVec bitStream vec.row1 #v4d; WriteVec bitStream vec.row2 #v4d; WriteVec bitStream vec.row3 #v4d; WriteVec bitStream vec.row4 #v4d;)
		#sphere: (WriteVec bitStream vec.emin #v3d; WriteVec bitStream vec.emax #v3d; WriteFloat bitStream vec.rad;)
	)
	if (bflags == #MD34) then WriteLong bitStream 0x00 #unsigned
)

fn WriteBlank bitStream count flags:#buffer ret:false =
(
	local bm = ftell bitStream
	local bcount = 0 -- buffer count

	if (flags != #blank) then
	(
		bcount = (16 - (mod count 16)) -- calculate buffer bytes
		if (mod count 16 == 0) then bcount -= 16 -- so we don't get rows of 16 buffer bytes
	)
	
	case flags of
	(
		#buffer: 
		(
			count = 0 -- don't let it write null data
		)
	)
	
	for i = 1 to (count as integer) do (WriteByte bitStream 0) -- write blank bytes
	for i = 1 to (bcount as integer) do (WriteByte bitStream 0xAA #unsigned) -- write buffer bytes
	if (ret != false) then fseek bitStream bm #seek_set
)

fn FloatToShort &val = -- Float to Short func
(
	if (val > 0) then (val -= 1) else (val += 1)
	val *= 32767.0
)

fn getSize type flags:0 =
(
	local size
	case type of
	(
		#REF: size = 0x0C
		#AREF: size = 0x20
		#MD33: size = 0x14
		#MD34: size = 0x18
		#MODL:
		(
			case flags of
			(
				23: size = 0x310
			)
		)
		#SEQS: size = 0x60
		#STC_: size = 0xCC
		#SDEV: size = getSize #AREF
		#EVNT: size = 0x68
		#STG_: size = 0x18
		#STS_: size = 0x1C
		#BONE: size = 0xA0
		#VERT: size = 0x20
		#U16_: size = 0x02
		#U32_: size = 0x04
		#I32_: size = 0x04
		#CHAR: size = 0x01
		#DIV_: size = 0x34
		#REGN: size = 0x24
		#BAT_: size = 0x0E
		#MSEC: size = 0x48
		#MATM: size = 0x08
		#MAT_: size = 0x10C
		#LAYR: size = 0x164
		#IREF: size = 0x40
	)
	
	return size
)

fn M3E_Name name flt ext:undefined =
(
	-- Filter name from path
	local mn = filterString name flt
	mn = mn[mn.count]
	local mext = substring mn (mn.count-3) (4)
	-- Remove extension if specified
	if (ext != undefined) then
	(
		if (matchPattern mext pattern:ext ignoreCase:true) then mn = substring mn 1 (mn.count-4)
	)
	return mn
)

-- ************************
-- M3 FUNCTIONS
-- ************************ 
fn M3E_Open fname flags =
(
	M3E_Reset_Globals()
	
	step = "Accessing Model File"
	
	echo ("Writing "+fname+"...")
	bstream = fopen fname flags
	if bstream==undefined then 
	(
		echo "File not found!"
		throw "File not found"
	)
)

fn M3E_Close bitStream =
(
	step = "Close"
	fclose bitStream
)

struct M3E_Tag -- Tag data
(
	dataID, ofsData, nData, flags,
	
	fn Write bitStream tag =
	(
		-- head to end of file
		fseek bitStream 0 #seek_end
		
		-- Fix Header info
		head.ofsTag = ftell bitStream
		head.nTag = tag.count
		
		-- write tags
		for i = 1 to tag.count do
		(
			t = tag[i]
			writeString bitStream t.dataID -- Write File ID
			fseek bitStream -1 #seek_cur
			writeLong bitStream t.ofsData #unsigned
			writeLong bitStream t.nData #unsigned
			writeLong bitStream t.flags #unsigned
		)
	),
	fn TagAppend bitStream nameID nData tagar flags:0 =
	(
		tag = M3E_Tag()
		local strID = nameID as string
		local tagID = "BLNK" -- Needs to be 4 chars for modification
		for i = 1 to 4 do
		(
			k = 5 - i
			tagID[i] = strID[k]
		)
		tag.dataID = toUpper tagID -- Uppercase
		tag.ofsData = ftell bitStream
		tag.nData = nData
		tag.flags = flags
		
		append tagar tag
	)
)

struct M3E_Ref
(
	entries, refid, flags = 0, data = #(),
	
	fn Write bitStream ref =
	(
		writeLong bitStream ref.entries #unsigned
		writeLong bitStream ref.refid #unsigned
		WriteLong bitStream ref.flags #unsigned
	)
)

struct M3E_AnimRef -- Animation Reference
(
	flags, animflags, animid,
	
	fn Write bitStream ar =
	(
		writeShort bitStream ar.flags #unsigned
		writeShort bitStream ar.animflags #unsigned
		writeLong bitStream ar.animid #unsigned
	),
	fn Default =
	(
		ar = M3E_AnimRef()
		ar.flags = 0
		ar.animflags = 0
		-- random uint32
		ar.animid = Random 1 100000000000
		
		return ar
	)
)

struct M3E_Header
(
	fileID, ofsTag, nTag, mref = M3E_Ref(),
	
	fn Write bitStream head =
	(
		step = "Write Header"

		fseek bstream 0 #seek_set -- move to beginning of file
		
		writeString bitStream head.fileID -- Write File ID
		fseek bitStream -1 #seek_cur -- writeString writes null terminator...so have to move cursor back one space
		writeLong bitStream head.ofsTag #unsigned
		writeLong bitStream head.nTag #unsigned
		M3E_Ref.Write bitStream head.mref
	),
	fn Default =
	(
		h = M3E_Header()
		-- Fixed file magic ID
		h.fileID = "43DM"
		-- Blank these for now, fix em up during tag writing
		h.ofsTag = 0
		h.nTag = 0
		-- Use fixed MODL ref...
		h.mref.entries = 1
		h.mref.refid = 1
		
		return h
	)
)

struct M3E_MODL
(
	name,  -- max model name
	versID, d1,
	SEQS, STC, STG, vec1, STS,
	bones, 
	vflags, verts, DIV, bonelu,
    ext1,
	matm, mat,
	IREF,
	
	fn Write bitStream modl vers = 
	(
		-- goto modl offset first
		fseek bitStream twrite[2].ofsData #seek_set
		
		-- Write MODL chunk
		WriteRef bitStream modl.name #CHAR
		WriteLong bitStream modl.versID #unsigned
		WriteRef bitStream modl.SEQS #SEQS flags:1
		WriteRef bitStream modl.STC #STC_ flags:4
		WriteRef bitStream modl.STG #STG_
		WriteVec bitStream modl.vec1 #v3d bflags:#MD34
		WriteRef bitStream modl.STS #STS_
		WriteRef bitStream modl.bones #BONE
		WriteLong bitStream modl.d1 #unsigned
		WriteLong bitStream modl.vflags #unsigned
		WriteRef bitStream modl.verts #VERT
		WriteRef bitStream modl.DIV #DIV_ flags:2
		WriteRef bitStream modl.bonelu #U16_ subtype:#uint16
		WriteVec bitStream modl.ext1 #sphere bflags:#MD34
		WriteBlank bitStream 0x84 flags:#blank
		WriteRef bitStream modl.MATM #MATM
		WriteRef bitStream modl.MAT #MAT_ flags:15
		WriteBlank bitStream 0xFC flags:#blank
		WriteRef bitStream modl.iref #IREF
	)
)

struct M3E_SEQS
(
	d1 = #(-1, -1), name, d2 = 0, maxframes, flags, d3 = #(0, 100, 1, 1, 100, 0), bndSphere = M3E_Sphere(), d4 = #(0,0),
	cstart, cend, -- Max variables
	
	fn Write bitStream ad =
	(
		WriteArray bitStream ad.d1 flags:#int32
		WriteRef bitStream ad.name #CHAR
		WriteLong bitStream ad.d2 #unsigned
		WriteLong bitStream ad.maxframes #unsigned
		WriteLong bitStream ad.flags #unsigned
		WriteArray bitStream ad.d3 flags:#uint32
		WriteVec bitStream ad.bndSphere #sphere
		WriteArray bitStream ad.d4 flags:#int32
	),
	fn Default =
	(
		a = M3E_SEQS()
		
		a.name = "Stand"
		a.maxframes = 10000
		a.flags = 0
		a.bndSphere = M3E_Sphere.GetBound cmesh[1]
		
		return a
	)
)

struct M3E_Event
(
	name = "Evt_SeqEnd", d1 = -1, s1 = -1, s2 = 0, mat1 = matrix(), d2 = #(4,0,0,0),
	
	fn Write bitStream evnt =
	(
		WriteRef bitStream evnt.name #CHAR
		WriteLong bitStream evnt.d1 #signed
		WriteShort bitStream evnt.s1 #signed
		WriteShort bitStream evnt.s2 #unsigned
		WriteVec bitStream evnt.mat1 #matrix bflags:#null
		WriteArray bitStream evnt.d2 flags:#uint32
	)
)

struct M3E_AnimBlock -- Sequence Data
(
	frames, flags, fend, keys,
	
	fn Write bitStream ablock sdtype =
	(
		WriteRef bitStream ablock.frames #I32_ subtype:#uint32
		WriteLong bitStream ablock.flags #unsigned
		WriteLong bitStream ablock.fend #unsigned
		WriteRef bitStream ablock.keys sdtype
	),
	fn DefaultEvnt =
	(
		a = M3E_AnimBlock()
		a.frames = #(9999)
		a.flags = 1
		a.fend = 10000
		a.keys = #(M3E_Event())
		
		return a
	)
)

struct M3E_Animoffs 
(
	aind = 0, sdind = 0,
	
	fn Write bitStream aoffs =
	(
		WriteShort bitStream aoffs.aind #unsigned
		WriteShort bitStream aoffs.sdind #unsigned
	)
)

struct M3E_STC --Sequence Transformations Collection (STC)
(
	name, d1, seqlu, stglu, animid, animref = M3E_Animoffs(), d2, sdata = #(),
	
	fn Write bitStream stc =
	(
		WriteRef bitStream stc.name #CHAR
		WriteLong bitStream stc.d1 #unsigned
		WriteShort bitStream stc.seqlu #unsigned
		WriteShort bitStream stc.stglu #unsigned
		WriteRef bitStream stc.animid #U32_ subtype:#uint32
		WriteRef bitStream stc.animref #U32_ subtype:#animind
		WriteLong bitStream stc.d2 #unsigned
		WriteRef bitStream stc.sdata[1] #SDEV
		for i = 2 to 13 do WriteRef bitStream stc.sdata[i] #null
	),
	fn Default =
	(
		s = M3E_STC()
		s.name = "Stand_full"
		s.d1 = 0
		s.seqlu = 0
		s.stglu = 0
		s.animid = #(Random 1 100000000000) -- must be array
		s.animref = #(M3E_Animoffs())
		s.d2 = 0
		-- initialize array
		s.sdata[13] = 0
		s.sdata[1] = #(M3E_AnimBlock.DefaultEvnt())
		for i = 2 to 13 do s.sdata[i] = #null
		
		return s
	)
)

struct M3E_STG
(
	name = "Stand", stcind = #(0),
	
	fn Write bitStream stg =
	(
		WriteRef bitStream stg.name #CHAR
		WriteRef bitStream stg.stcind #U32_ subtype:#uint32
	)
)

struct M3E_STS
(
	animid, d1 = #(-1,-1,-1), s1 = -1, s2 = 0,
	
	fn Write bitStream sts =
	(
		WriteRef bitStream sts.animid #U32_ subtype:#uint32
		WriteArray bitStream sts.d1 flags:#int32
		WriteShort bitStream sts.s1 #signed
		WriteShort bitStream sts.s2 #unsigned
	),
	fn Default modl =
	(
		s = M3E_STS()
		
		s.animid = #(modl.stc[1].animid[1])
		
		return s
	)
)

struct M3E_Bone
(
	d1, name, flags, parent, s1, 
	transar, pos1, pos2, d2,  
	rotar, rot1, rot2, d3, 
	scalear, scale1, scale2, 
	d4, ar1, unk1, unk2, d5,
	
	fn Write bitStream bone =
	(
		WriteLong bitStream bone.d1 #signed
		WriteRef bitStream bone.name #CHAR
		WriteLong bitStream bone.flags #unsigned
		WriteShort bitStream bone.parent #signed
		WriteShort bitStream bone.s1 #unsigned
		M3E_AnimRef.Write bitStream bone.transar
		WriteVec bitStream bone.pos1 #v3d
		WriteVec bitStream bone.pos2 #v3d
		WriteLong bitStream bone.d2 #unsigned
		M3E_AnimRef.Write bitStream bone.rotar
		WriteVec bitStream bone.rot1 #v4d
		WriteVec bitStream bone.rot2 #v4d
		WriteLong bitStream bone.d3 #uint32
		M3E_AnimRef.Write bitStream bone.scalear
		WriteVec bitStream bone.scale1 #v3d
		WriteVec bitStream bone.scale2 #v3d
		WriteLong bitStream bone.d4 #unsigned
		M3E_AnimRef.Write bitStream bone.ar1
		WriteLong bitStream bone.unk1 #unsigned
		WriteLong bitStream bone.unk2 #unsigned
		WriteLong bitStream bone.d5 #unsigned
	),
	fn Default =
	(
		b = M3E_Bone()
		
		b.d1 = -1
		b.name = "Main"
		b.flags = 2048
		b.parent = -1
		b.s1 = 0
		b.transar = M3E_AnimRef.Default()
		b.pos1 = [0,0,0]
		b.pos2 = [0,0,0]
		b.d2 = 0
		b.rotar = M3E_AnimRef.Default()
		b.rot1 = [0,0,0,1]
		b.rot2 = [0,0,0,1]
		b.d3 = 0
		b.scalear = M3E_AnimRef.Default()
		b.scale1 = [1,1,1]
		b.scale2 = [1,1,1]
		b.d4 = 0
		b.ar1 = M3E_AnimRef.Default()
		b.unk1 = 1
		b.unk2 = 1
		b.d5 = 0
		
		return b
	)		
)

struct M3E_Vertex
(
	pos = [0,0,0], bw = #(0,0,0,0), bi = #(0,0,0,0), normal = #(0,0,0,0), uv = #(0,0), d1 = 0, d2 = 0,
	faceid, faceref = #(), origvertindx, origtvertindx, -- for script purposes only
	
	fn Write bitStream vert =
	(
		WriteVec bitStream vert.pos #v3d
		WriteArray bitStream vert.bw flags:#uint8
		WriteArray bitStream vert.bi flags:#uint8
		WriteArray bitStream vert.normal flags:#uint8
		WriteArray bitStream vert.uv flags:#int16
		--WriteLong bitStream vert.d1 #unsigned
		WriteLong bitStream vert.d2 #unsigned
	),
	fn buildVertices smesh =
	(
		local varray = #()
		for i = 1 to smesh.count do
		(
			sm = smesh[i]
					
			for i = 1 to sm.verts.count do
			(
				v = sm.verts[i]
				local v, vpos, vnorm, vuv
				local cnorm, cuv
				
				-- set fixed default weight
				v.bw[1] = 255
				
				-- compress normals
				cnorm = #(0,0,0,0)
				for i = 1 to 3 do
				(
					local cn = 1 + v.normal[i]
					cn = (cn / 2) * 255
					cnorm[i] = cn as integer
				)
				v.normal = cnorm
				
				-- compress UV's
				cuv = #(0,0)
				if (g_flipuvy) then v.uv[2] = 1 - v.uv[2]
				for i = 1 to 2 do
				(
					
					cuv[i] = (v.uv[i] * 2046) as integer
				)
				
				v.uv = cuv
				
				append varray v
			)
		)
		
		return varray
	)
)

struct M3E_REGN
(
	s1 = #(0,0), indvert, nvert, 
	indface, nface, 
	bcount = 1, indbonelu = 0, nbonelu = 1,
	s2 = 0, b1 = #(1,1), s3 = 0,
	verts = #(), faces = #(),

	fn Write bitStream regn =
	(
		WriteArray bitStream regn.s1 flags:#uint32
		WriteLong bitStream regn.indvert #unsigned
		WriteLong bitStream regn.nvert #unsigned
		WriteLong bitStream regn.indface #unsigned
		WriteLong bitStream regn.nface #unsigned
		WriteShort bitStream regn.bcount #unsigned
		WriteShort bitStream regn.indbonelu #unsigned
		WriteShort bitStream regn.nbonelu #unsigned
		WriteShort bitStream regn.s2 #unsigned
		WriteArray bitStream regn.b1 flags:#uint8
		WriteShort bitStream regn.s3 #unsigned
	)
)

struct M3E_BAT
(
	d1 = 0, subid, d2 = 0, matid, s1 = -1,
	
	fn Write bitStream bat =
	(
		WriteLong bitStream bat.d1 #unsigned
		WriteShort bitStream bat.subid #unsigned
		WriteLong bitStream bat.d2 #unsigned
		WriteShort bitStream bat.matid #unsigned
		WriteShort bitstream bat.s1 #signed
	)
)

struct M3E_MSEC
(
	d1 = 0, ar1 = M3E_AnimRef.Default(), ext1 = M3E_Sphere(), d8 = #(0,0,0,0,0,0,0,0),
	
	fn Write bitStream msec =
	(
		WriteLong bitStream msec.d1 #unsigned
		M3E_AnimRef.Write bitStream msec.ar1
		WriteVec bitStream msec.ext1 #sphere
		WriteArray bitStream msec.d8 flags:#uint32
	)
)

struct M3E_DIV -- Model divisions
(
	faces = #(), regn = #(), bat = #(), msec = #(), d1 = 1,
	
	fn Write bitStream div =
	(
		WriteRef bitStream div.faces #U16_ subtype:#uint16
		WriteRef bitStream div.regn #REGN flags:3
		WriteRef bitStream div.bat #BAT_ flags:1
		WriteRef bitStream div.msec #MSEC flags:1
		WriteLong bitStream div.d1 #unsigned
	)
)

struct M3E_Subref -- Submesh refrences
(
	subid, matid
)

struct M3E_MATM
(
	nmat = 1, matind,
	
	fn Write bitStream matm =
	(
		WriteLong bitStream matm.nmat #unsigned
		WriteLong bitStream matm.matind #unsigned
	),
	fn Read mats =
	(
		marray = #()
		for i = 1 to mats.count do
		(
			m = M3E_MATM()
			m.matind = mats[i] - 1
			
			append marray m
		)
		
		return marray
	)
)

struct M3E_LAYR --more work to be done at a later point
(
	name, f1 = 1, f2 = 1, flags = 236,
	
	fn Write bitStream layr =
	(
		WriteLong bitStream 0 #unsigned
		WriteRef bitStream layr.name #CHAR
		WriteBlank bitStream 0x14 flags:#blank
		WriteLong bitStream layr.flags #unsigned
		WriteBlank bitStream 0x10 flags:#blank
		WriteFloat bitStream layr.f1
		WriteFloat bitStream layr.f2
		WriteBlank bitStream 0x1C flags:#blank
		WriteLong bitStream -1 #signed
		WriteBlank bitStream 0xEC flags:#blank
		WriteLong bitStream -1 #signed
		WriteBlank bitStream 0x14 flags:#blank
	)
)

struct M3E_Mat
(
	name, flags = 540676, layrs = #(),
	
	fn Write bitStream mat =
	(
		WriteRef bitStream mat.name #CHAR
		WriteBlank bitStream 0x04 flags:#blank
		WriteLong bitStream mat.flags #unsigned
		WriteBlank bitStream 0x20 flags:#blank
		for i = 1 to 13 do WriteRef bitStream mat.layrs[i] #LAYR flags:22
		WriteBlank bitStream 0x3C flags:#blank
	)
)

-- ****************
--  EXPORT CODE
-- ****************
---------------------
-- SCENE CODE --
---------------------
-- borrowed code from der_ton's MD5 exporter
fn checkforvertex vertindx tvertindx verts =
(
	for i=1 to verts.count do
	(
		if (vertindx == verts[i].origvertindx and tvertindx == verts[i].origtvertindx) do
		(
			return i
		)
	)
	return undefined
)

fn buildFaces obj triindx regn =
(
	vertindices = getface obj triindx
	if (meshop.getMapSupport obj 1) then
		tvertindices = meshop.getmapface obj g_mapchannel triindx
	else
		tvertindices = [0,0,0]
	vertindices = #(vertindices.x as integer, vertindices.y as integer, vertindices.z as integer)
	tvertindices = #(tvertindices.x as integer, tvertindices.y as integer, tvertindices.z as integer)
	m3vertindices = #()
	for i=1 to 3 do
	(
		m3vertindices[i] = checkforvertex vertindices[i] tvertindices[i] regn.verts
		if (m3vertindices[i] == undefined) do
		(
			vind = vertindices[i]
			tvind = tvertindices[i]
			v = M3E_Vertex()
			v.pos 	= getVert obj vind
			v.normal	= getNormal obj vind
			if (meshop.getMapSupport obj g_mapchannel) then
				v.uv = meshop.getmapvert obj g_mapchannel tvind
			else uv = [0,0,0]
			v.origvertindx = vind
			v.origtvertindx = tvind
			append regn.verts v
			m3vertindices[i] = regn.verts.count
		)
	)
	for i = 1 to 3 do
	(
		append regn.faces (m3vertindices[i] - 1)
	)
)

fn getSubmesh mesh =
(
	div = M3E_DIV()
	for i = 1 to mesh.count do
	(
		m = mesh[i]
		
		-- cycle through faces, gather data
		for k = 1 to m.numfaces do
		(
			-- #MATERIALS/SUBMESHES
			-- Gather Used Materials
			local matid = getfacematid m k
			local found = false
			local j = 1
			
			while (j <= usedMats.count and found == false) do
			(
				if usedMats[j] == matid then 
				(
					found = true
				)
				j+=1
			)
			if found == false then
			(
				append usedMats matid
			)
		)
		
		for i = 1 to usedMats.count do
		(
			-- Create Submeshes
			r = M3E_REGN()
			b = M3E_BAT()
			b.subid = i - 1
			b.matid = usedMats[i] - 1

			append div.regn r
			append div.bat b
		)
		
		-- collect faces into the submeshes, according to the material the faces use
		for k=1 to m.numfaces do
		(
			matid = getfacematid m k
			j=1 
			local found = false
			while (j<usedMats.count and found == false) do
			(
				if usedMats[j] == matid then found = true
				else j+=1
			)
			
			buildFaces m k div.regn[j]
		)
	)
	
	--echo ("Numsubs: "+div.regn.count as string)
	--echo ("------------------------------")
	
	return div
)

fn buildDIV div =
(
	-- run through submeshes
	for i = 1 to div.regn.count do
	(
		r = div.regn[i]
		
		r.nvert = r.verts.count
		r.nface = r.faces.count
		if (i > 1) then
		(
			prevr = div.regn[i-1]
			r.indvert = prevr.indvert + prevr.nvert
			r.indface = prevr.indface + prevr.nface
		)
		else
		(
			r.indvert = 0
			r.indface = 0
		)
		
		-- build single faces list
		join div.faces r.faces
	)
	
	div.msec = #(M3E_MSEC())
)

fn getMaterials mesh mats =
(
	mtlarray = #()
	for i = 1 to mesh.count do
	(
		m = mesh[i]
		
		for h = 1 to mats.count do
		(
			mtl = M3E_Mat()
			matid = mats[h]
			
			if ((classof m.material as string) != "Multimaterial") then
			(
				messagebox "Mesh material must be multimaterial!"
				throw "Mesh material must be multimaterial!"
			)
			
			mmat = m.material[matid]
			mtl.name = mmat.name
			for k = 1 to 13 do
			(
				local layr = M3E_LAYR()
				layr.name = ""
				case k of
				(
					1:
					(
						if (mmat.diffuseMapEnable) then 
						(
							layr.name = texPath + (M3E_Name mmat.diffuseMap.bitmap.filename "\\")
							--echo ("diffuseMap: "+layr.name)
						)
					)
					3:
					(
						if (mmat.specularMapEnable) then 
						(
							layr.name = texPath + (M3E_Name mmat.specularMap.bitmap.filename "\\")
							--echo ("specularMap: "+layr.name)
						)
					)
					4:
					(
						if (mmat.selfIllumMapEnable) then 
						(
							layr.name = texPath + (M3E_Name mmat.selfIllumMap.bitmap.filename "\\")
							--echo ("emissiveMap: "+layr.name)
						)
					)
					10: 
					(
						if (mmat.bumpMapEnable) then 
						(
							layr.name = texPath + (M3E_Name mmat.bumpMap.bitmap.filename "\\") 
						)
						--echo ("normalMap: "+layr.name)
					)
				)
				
				append mtl.layrs #(layr)
			)
			append mtlarray mtl
		)
	)
	
	return mtlarray
)

---------------------
-- M3 HELPERS  --
---------------------

fn WriteRef bitStream data reftype subtype:#none flags:0  =
(
	local ref = M3E_Ref()
	-- Make sure it isn't a null reference...
	if (data != #null) then
	(
		-- MODL bookmark
		local bm = ftell bitStream
		
		-- head to end of file to write reference data
		fseek bitStream 0 #seek_end
		
		-- Calculate chunk size
		-- need special VERT clause
		local byte_size = (getSize reftype) * data.count
		case reftype of
		(
			#VERT: (M3E_Tag.TagAppend bitStream #U8__ (byte_size) twrite flags:flags; ref.entries = byte_size;)
			#CHAR: (byte_size += 1; M3E_Tag.TagAppend bitStream #CHAR (byte_size) twrite flags:flags)
			default: M3E_Tag.TagAppend bitStream reftype (data.count) twrite flags:flags
		)
		WriteBlank bitStream byte_size flags:#none ret:true
		ref.refid = twrite.count - 1 -- convert to base 0
		
		-- ** REF DEBUG **
		/* 
		echo ("reftype: "+reftype as string)
		echo ("refid: "+ref.refid as string)
		echo ("data.count: "+data.count as string)
		echo ("byte_size: "+byte_size as string)
		echo ("ftell: "+(ftell bitstream) as string)
		*/
		
		if (reftype != #CHAR) then
		(
			for i = 1 to data.count do
			(
				d = data[i]
				case reftype of
				(
					#SEQS: M3E_SEQS.Write bitStream d
					#STC_: M3E_STC.Write bitStream d
					#SDEV: M3E_AnimBlock.Write bitStream d #EVNT
					#EVNT: M3E_Event.Write bitStream d
					#STG_: M3E_STG.Write bitStream d
					#STS_: M3E_STS.Write bitStream d
					#BONE: M3E_Bone.Write bitStream d
					#VERT: M3E_Vertex.Write bitStream d
					#REGN: M3E_REGN.Write bitStream d
					#BAT_: M3E_BAT.Write bitStream d
					#MSEC: M3E_MSEC.Write bitStream d
					#MATM: M3E_MATM.Write bitStream d
					#MAT_: M3E_MAT.Write bitStream d
					#LAYR: M3E_LAYR.Write bitStream d
					#DIV_: M3E_DIV.Write bitStream d
					#IREF: WriteVec bitStream d #matrix
					#U16_: WriteArray bitStream d flags:subtype abool:false
					#U32_: WriteArray bitStream d flags:subtype abool:false
					#I32_: WriteArray bitStream d flags:subtype abool:false
				)
			)
		)
		else
		(
			WriteString bitStream data
			-- fix tag entry
			twrite[twrite.count].nData = byte_size
			ref.entries = byte_size
		)
		
		if (ref.entries == undefined) then ref.entries = data.count
		-- return to MODL bookmark
		fseek bitStream bm #seek_set
	)
	else
	(
		-- setup blank reference
		ref.entries = 0
		ref.refid = 0
	)
	
	-- write reference 
	M3E_Ref.Write bitStream ref
)

fn WriteArray bitStream data flags:#uint32 reftype:#null abool:true  =
(
	-- if not array...
	if (abool != true) then icount = 1 else icount = data.count
	for i = 1 to icount do
	(
		local d
		if (abool != true) then d = data else d = data[i]
		case flags of
		(
			#Ref: WriteRef bitStream d reftype
			#uint32: writeLong bitStream d #unsigned
			#int32: WriteLong bitStream d #signed
			#uint16: writeShort bitStream d #unsigned
			#int16: writeShort bitStream d #signed
			#uint8: writeByte bitStream d #unsigned
			#int8: writeByte bitStream d #signed
			#float: writeFloat bitStream d
			#v3d: WriteVec bitStream d #v3d
			#animind: M3E_AnimOffs.Write bitStream d
		)
	)
)

fn M3E_GetMesh =
(
	cm = #()
	Select geometry
	local sel = getCurrentSelection() -- store Mesh!
	
	max modify mode
	set coordsys world
	
	-- Mesh sort, deselect unsuitable meshes
	--for i = 1 to sel.count do
	if (sel.count < 2) then
	(
		m = sel[1]
		if ((classof m.baseObject as string) != "Editable_mesh") then
		(
			--echo ("#Warning! "+w_mesh as string+" is not an Editable_mesh class!")
			--echo ("Class of object: "+classof w_mesh.baseObject as string)
			deselect m
		)
		else
		(
			if (m.isHidden == true) or (m.isFrozen == true) then -- do not export hidden/frozen geometry, will make this a selectable option!
			(
				deselect w_mesh
			)
			else
			(
				-- append to mesh list
				append cm m
			)
		)
	)
	else
	(
		messagebox "Only one mesh currently allowed!"
		throw "Only one mesh currently allowed!"
	)

	csel = GetCurrentSelection()
	m = csel[1]
	if cm.count < 1 then
	(
		messagebox "No Editable Meshes Found!"
		throw "No Editable Meshes Found!"
	)
	
	return cm
)

fn M3E_GetMODL =
(
	m = M3E_MODL()
	
	--m.name = "Test.max"
	m.name = "Prototype.max"
	m.versID = 1576275
	m.SEQS = #(M3E_SEQS.Default())
	m.STC = #(M3E_STC.Default())
	m.STG = #(M3E_STG())
	m.vec1 = [0,0,0]
	m.STS = #(M3E_STS.Default m)
	m.bones = #(M3E_Bone.Default())
	m.d1 = 1
	-- grab submeshes first
	m.DIV = #(getSubmesh cmesh)
	buildDIV m.DIV[1]
	
	-- build verts based on submesh list
	m.vflags = 25297021
	m.verts = M3E_Vertex.buildVertices m.DIV[1].regn
	m.bonelu = #(0)
	m.ext1 = M3E_Sphere.GetBound cmesh[1] -- use only first mesh for bounds for the moment...
	m.matm = M3E_MATM.Read usedMats
	m.mat = getMaterials cmesh usedMats
	m.iref = #(matrix())
	
	return m
)


fn M3E_Write_Start =
(
	step = "Start Model Write"
	scrStart = TimeStamp() -- start tracking script processing time
	-- Create Header Tag, write Header blank bytes
	head = M3E_Header.Default()
	M3E_Tag.TagAppend bstream #MD34 1 twrite flags:11
	WriteBlank bstream (getSize #MD34) flags:#none
	
	-- Create MODL tag, write MODL blank bytes
	-- using model version 23
	M3E_Tag.TagAppend bstream #MODL 1 twrite flags:23
	WriteBlank bstream (getSize #MODL flags:23) flags:#none
	
	-- Get viable Meshes
	cmesh = M3E_GetMesh()
	-- Gather MODL data
	-- assign global, will update this structure throughout export
	mwrite = M3E_GetMODL()
	M3E_MODL.Write bstream mwrite 23
)

fn M3E_Main file =
(
	M3E_Open file "wb"
	M3E_Write_Start()
	
	-- Write Tags second last...
	M3E_Tag.Write bstream twrite
	
	-- ... and write Header last
	M3E_Header.Write bstream head
	M3E_Close bstream
	-- Calculate Script Execution time
	scrEnd = TimeStamp()
	format "Export took % seconds\n" ((scrEnd - scrStart) / 1000.0)
	echo "====== Export Successful ======\n"
	
	setCommandPanelTaskMode #utility
)

-- *********************************
--  USER INTERFACE FUNCTIONS
-- *********************************
-- Glob UI Funcs
fn M3E_GetFile ftypes =
(
	local ret = getSaveFileName types:ftypes
	if ret != undefined then return ret
)

utility M3E_export "M3 - Export"
(
	group "Texture Options"
	(
		label	lblTexPath "Texture Path:"
		edittext tTexPath text:texPath
		spinner spnMatID "Material ID:" range:[1,21,1] type:#integer
		checkbox chkFlipUVY "Flip UV Y-coord" checked:(g_flipuvy)
	)
	
	button bExport "Export" height:35 width:100
	
	on bExport pressed do
	(
		local fname = getSaveFileName types:"M3 model (*.m3)|*.m3|All Files|*.*|"
		if fname != undefined then
		(
			g_mapchannel = spnMatID.value
			g_flipuvy = chkFlipUVY.checked
			texPath = tTexPath.text
			M3E_Main(fname)
		)
	)
)